searchNodes=[{"doc":"This module provides the high level interface for interacting with topics. Primarily for subscribing (and unsubscribing), but also for sending requests. After subscribing, a client will initially receive a {:reset, ref, value} message, and then subsequent {:updates, ref, updates} messages when the value of the topic changes, where updates is a list with each item being one of: {:set, path, value} : the value has been set at the path . {:unset, path, key} : the key has been unset from the object at the path . {:insert, path, index, values} : the values have been inserted into the array at the path . {:delete, path, index, count} : count values have been deleted from the array at the path , from the position index . A client can interact directly with a topic by executing actions (which returns a result), or by notifying (without waiting for a result). These are analogous to GenServer.call/3 and GenServer.cast/2 . Be aware that a topic is blocked while processing a request.","ref":"Topical.html","title":"Topical","type":"module"},{"doc":"Captures the state of the topic (in the specified registry ) without subscribing. Example Topical . capture ( MyApp.Topical , &quot;lists/foo&quot; ) # =&gt; {:ok, %{items: %{}, order: []}}","ref":"Topical.html#capture/3","title":"Topical.capture/3","type":"function"},{"doc":"Returns a specification to start a Topical registry under a supervisor.","ref":"Topical.html#child_spec/1","title":"Topical.child_spec/1","type":"function"},{"doc":"Executes an action in a topic . Example Topical . execute ( MyApp.Topical , &quot;lists/foo&quot; , &quot;add_item&quot; , { &quot;Test&quot; , false } ) #=&gt; {:ok, &quot;item123&quot;}","ref":"Topical.html#execute/5","title":"Topical.execute/5","type":"function"},{"doc":"Send a notification to a registry. This is similar to execute/4 , except no result is waited for. Example Topical . notify ( MyApp.Topical , &quot;lists/foo&quot; , &quot;update_done&quot; , { &quot;item123&quot; , true } ) #=&gt; :ok","ref":"Topical.html#notify/5","title":"Topical.notify/5","type":"function"},{"doc":"Subscribes to the specified topic (in the specified registry ). Returns {:ok, ref} , where the ref is a reference to the subscription. The pid will be send messages, as described above. Example Topical . subscribe ( MyApp.Topical , &quot;lists/foo&quot; , self ( ) ) #=&gt; {:ok, #Reference&lt;0.4021726225.4145020932.239110&gt;}","ref":"Topical.html#subscribe/4","title":"Topical.subscribe/4","type":"function"},{"doc":"Unsubscribes from a topic (in the specified registry ). Example Topical . unsubscribe ( MyApp.Topical , &quot;lists/foo&quot; , ref )","ref":"Topical.html#unsubscribe/3","title":"Topical.unsubscribe/3","type":"function"},{"doc":"A REST-ish handler adapter for a Cowboy web server. Options registry - The name of the Topical registry. Required. init - A function called before starting/capturing the topic, passed the request. The function must return {:ok, context} for the connection to be accepted. This context is then passed to topics. Example :cowboy_router . compile ( [ { :_ , [ # ... { &quot;/topics/[...]&quot; , RestHandler , registry : MyApp.Topical } ] } ] )","ref":"Topical.Adapters.Cowboy.RestHandler.html","title":"Topical.Adapters.Cowboy.RestHandler","type":"module"},{"doc":"A Websocket handler adapter for a Cowboy web server. Options registry - The name of the Topical registry. Required. init - A function called before upgrading the connection, which is passed the request. The function must return {:ok, context} for the connection to be accepted. This context is then passed to topics. Example :cowboy_router . compile ( [ { :_ , [ # ... { &quot;/socket&quot; , WebsocketHandler , registry : MyApp.Topical } ] } ] )","ref":"Topical.Adapters.Cowboy.WebsocketHandler.html","title":"Topical.Adapters.Cowboy.WebsocketHandler","type":"module"},{"doc":"This module provides functions for instantiating and manipulating topic state. The state of a topic is composed of a value , observed by subscribed clients, and also hidden internal state, which can be used to share state between calls. The value must only be manipulated by the helper functions (which track the individual updates). The hidden state can be modified directly. This module also contains a macro - use -ing it sets up a topic server: defmodule MyApp.Topics.List do use Topical.Topic , route : &quot;lists/:list_id&quot; # Initialise the topic def init ( params ) do list_id = Keyword . fetch! ( params , :list_id ) value = %{ items : %{ } , order : [ ] } # exposed &#39;value&#39; of the topic state = %{ list_id : list_id } # hidden server state topic = Topic . new ( value , state ) { :ok , topic } end # Optionally, handle subscribe def handle_subscribe ( topic , _context ) do { :ok , topic } end # Optionally, handle unsubscribe def handle_unsubscribe ( topic , _context ) do { :ok , topic } end # Optionally, handle capture def handle_capture ( topic , _context ) do { :ok , topic } end # Optionally, handle execution of an action def handle_execute ( &quot;add_item&quot; , { text } , topic , _context ) do id = Integer . to_string ( :erlang . system_time ( ) ) # Update the topic by putting the item in &#39;items&#39;, and appending the id to &#39;order&#39; topic = topic |&gt; Topic . set ( [ :items , id ] , %{ text : text , done : false } ) |&gt; Topic . insert ( [ :order ] , id ) # Return the result (the &#39;id&#39;), and the updated topic { :ok , id , topic } end # Optionally, handle a notification (an action without a result) def handle_notify ( &quot;update_text&quot; , { id , text } , topic , _context ) do topic = Topic . set ( topic , [ :items , id , :text ] , text ) { :ok , topic } end # Optionally, handle Erlang messages def handle_info ( { :done , id } , topic ) do topic = Topic . set ( topic , [ :items , id , :done ] , true ) { :ok , topic } end # Optionally, handle the topic being terminated (e.g., once clients have disconnected) def terminate ( _reason , topic ) do # ... end end","ref":"Topical.Topic.html","title":"Topical.Topic","type":"module"},{"doc":"Updates the topic be deleting count values from the array at the path , from the index . %{ foo : %{ bar : [ 1 , 2 , 3 , 4 ] } } |&gt; Topic . new ( ) |&gt; Topic . delete ( [ :foo , :bar ] , 2 ) |&gt; Map . fetch! ( :value ) #=&gt; %{foo: %{bar: [1, 2, 4]}}","ref":"Topical.Topic.html#delete/4","title":"Topical.Topic.delete/4","type":"function"},{"doc":"Updates the topic by inserting the specified values into the array at the path , at the index (or append them to the end, if no index is specified). If value is a list, all the items will be added (to add a list as a single item, wrap it in a list). %{ foo : %{ bar : [ 1 , 4 ] } } |&gt; Topic . new ( ) |&gt; Topic . insert ( [ :foo , :bar ] , 1 , [ 2 , 3 ] ) |&gt; Map . fetch! ( :value ) #=&gt; %{foo: %{bar: [1, 2, 3, 4]}}","ref":"Topical.Topic.html#insert/4","title":"Topical.Topic.insert/4","type":"function"},{"doc":"Instantiates a new instance of a topic. value is the initial value of the client-visible state. state is the 'hidden' internal state.","ref":"Topical.Topic.html#new/2","title":"Topical.Topic.new/2","type":"function"},{"doc":"Updates the topic by setting the value at the path . %{ foo : %{ bar : 2 } } |&gt; Topic . new ( ) |&gt; Topic . set ( [ :foo , :bar ] , 3 ) |&gt; Map . fetch! ( :value ) #=&gt; %{foo: %{bar: 3}}","ref":"Topical.Topic.html#set/3","title":"Topical.Topic.set/3","type":"function"},{"doc":"Updates the topic by unsetting the key of the object at the path . %{ foo : %{ bar : 2 } } |&gt; Topic . new ( ) |&gt; Topic . unset ( [ :foo ] , :bar ) |&gt; Map . fetch! ( :value ) #=&gt; %{foo: %{}}","ref":"Topical.Topic.html#unset/3","title":"Topical.Topic.unset/3","type":"function"},{"doc":"This module defines the bahaviour of a topic. See Topical.Topic for a usage example.","ref":"Topical.Topic.Server.html","title":"Topical.Topic.Server","type":"behaviour"},{"doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Topical.Topic.Server.html#child_spec/1","title":"Topical.Topic.Server.child_spec/1","type":"function"},{"doc":"Invoked before state is captured (after initialisation). This callback is optional.","ref":"Topical.Topic.Server.html#c:handle_capture/2","title":"Topical.Topic.Server.handle_capture/2","type":"callback"},{"doc":"Invoked when a client has executed an action. This callback is optional. If one is not implemented, the topic will fail if an action is executed.","ref":"Topical.Topic.Server.html#c:handle_execute/4","title":"Topical.Topic.Server.handle_execute/4","type":"callback"},{"doc":"Invoked to handle other messages. This callback is optional.","ref":"Topical.Topic.Server.html#c:handle_info/2","title":"Topical.Topic.Server.handle_info/2","type":"callback"},{"doc":"Invoked when a client has sent a notification. This callback is optional. If one is not implemented, the topic will fail if a notification is received.","ref":"Topical.Topic.Server.html#c:handle_notify/4","title":"Topical.Topic.Server.handle_notify/4","type":"callback"},{"doc":"Invoked before a client subscribes (but after initialisation). This callback can be used to update the topic, for example (in combination with handle_unsubscribe ) to track connected users. This callback is optional.","ref":"Topical.Topic.Server.html#c:handle_subscribe/2","title":"Topical.Topic.Server.handle_subscribe/2","type":"callback"},{"doc":"Invoked after a client unsubscribes (either explicitly or because the process dies). This callback is optional.","ref":"Topical.Topic.Server.html#c:handle_unsubscribe/2","title":"Topical.Topic.Server.handle_unsubscribe/2","type":"callback"},{"doc":"Invoked when the topic is started to get the initial state. params are the values associated with the placeholders in the route.","ref":"Topical.Topic.Server.html#c:init/1","title":"Topical.Topic.Server.init/1","type":"callback"},{"doc":"Starts a topic server process linked to the current process. Options :module - the module that implements the topic behaviour. :init_arg - the argument passed to the topic's init callback.","ref":"Topical.Topic.Server.html#start_link/1","title":"Topical.Topic.Server.start_link/1","type":"function"},{"doc":"Invoked when a topic has been stopped. This callback is optional.","ref":"Topical.Topic.Server.html#c:terminate/2","title":"Topical.Topic.Server.terminate/2","type":"callback"},{"doc":"Topical is a library for defining and serving topics . A topic is a value that can be efficiently observed in real-time by connected clients. The server-side implementation takes care of defining how to intialise the state, and then how to keep it updated. Multiple clients can subscribe to an instance of a topic - instances are identifier by a path, which matches the route defined by a topic. Multiple instances of a topic can exist by using route placeholders. Topical takes care of starting topic instances as needed, and stopping them once all clients have disconnected.","ref":"introduction.html","title":"Introduction","type":"extras"},{"doc":"Topics are implemented in Elixir, and behave somewhat like GenServers - the main difference being that their state is easily observable by clients. The main benefit comes from using the JavaScript (and React) client (and Cowboy WebSocket adapter) to observe the state, but it's also possible to use the Elixir API. There's a certain amount of flexibility in how topics are implemented. They can be be used to track ephemeral state - for example user presence, or cursor positions. They can be backed by simple mechanisms like a file or file-based database. Or they can sit in front of an RDBMS (e.g., utilising Postgres notification channels), providing real-time cached views. Or they can be used to implement the event sourcing pattern. Another use case is for implementing an incremental cache: the state of the cache is setup in the topic initalisation and then kept up to date as necessary, without having to refresh from scratch. Clients accessing the cache will be able to read up-to-date state directly from memory.","ref":"introduction.html#uses","title":"Introduction - Uses","type":"extras"},{"doc":"First you need to implement a topic definition. For example, a todo list topic might track todo items and their order, handle requests from clients to add new items or update the text of existing items, and handle Erlang messages from a separate process indicating when items are done: defmodule MyApp.Topics.List do use Topical.Topic , route : &quot;lists/:list_id&quot; # Initialise the topic def init ( params ) do list_id = Keyword . fetch! ( params , :list_id ) value = %{ items : %{ } , order : [ ] } # exposed &#39;value&#39; of the topic state = %{ list_id : list_id , last_item_id : 0 } # hidden server state topic = Topic . new ( value , state ) { :ok , topic } end # Optionally, handle execution of an action def handle_execute ( &quot;add_item&quot; , { text } , topic , _context ) do { id , topic } = generate_item_id ( topic ) # Update the topic by putting the item in &#39;items&#39;, and appending the id to &#39;order&#39; topic = topic |&gt; Topic . set ( [ :items , id ] , %{ text : text , done : false } ) |&gt; Topic . insert ( [ :order ] , id ) # Return the result (the &#39;id&#39;), and the updated topic { :ok , id , topic } end defp generate_item_id ( topic ) do id = topic . state . last_item_id + 1 topic = put_in ( topic [ :state ] [ :last_item_id ] , id ) { Integer . to_string ( id ) , topic } end # Optionally, handle a notification (an action without a result) def handle_notify ( &quot;update_text&quot; , { id , text } , topic ) do topic = Topic . set ( topic , [ :items , id , :text ] , text ) { :ok , topic } end # Optionally, handle Erlang messages def handle_info ( { :done , id } , topic ) do topic = Topic . set ( topic , [ :items , id , :done ] , true ) { :ok , topic } end end Then add a Topical registry to your application supervision tree, referencing the topic: defmodule MyApp.Application do use Application @impl true def start ( _type , _args ) do children = [ # ... { Topical , name : MyApp.Topical , topics : [ MyApp.Topics.List , ... ] } , ] Supervisor . start_link ( children , ... ) end end At this point you should be able to subscribe to the topic: { :ok , ref } = Topical . subscribe ( MyApp.Topical , &quot;lists/foo&quot; , self ( ) ) After subscribing, the process will be sent an initial {:reset, ref, value} message (where ref is the subscription reference returned from subscribe ), and then subsequent {:updates, ref, updates} messages, where updates is a list of updates, each taking the form: {:set, path, value} : the value has been set at the path . {:unset, path, key} : the key has been unset from the object at the path . {:insert, path, index, values} : the values have been inserted into the array at the path . {:delete, path, index, count} : count values have been deleted from the array at the path , from the position index . (You can receive any waiting messages on an IEx shell with: receive do x -&gt; x after 0 -&gt; nil end .) To execute an action: { :ok , item_id } = Topical . execute ( MyApp.Topical , &quot;lists/foo&quot; , &quot;add_item&quot; , { &quot;Test item&quot; , false } ) To unsubscribe: Topical . unsubscribe ( MyApp.Topical , &quot;lists/foo&quot; , ref ) However, rather than using the API from Elixir, you may wish to set up the Cowboy adapter , and then use the JavaScript client .","ref":"getting-started.html","title":"Getting started","type":"extras"},{"doc":"Rather than interacting with with the Topical API directly, you can simply expose Topical from your web server. If you're using Cowboy, this can be easily achieved using the Cowboy WebSocket handler adapter: :cowboy_router . compile ( [ { :_ , [ # ... { &quot;/socket&quot; , Topical.Adapters.Cowboy.WebsocketHandler , registry : Todo.Registry } ] } ] ) The Topical registry must be specified.","ref":"cowboy-adapter.html","title":"Cowboy adapter","type":"extras"},{"doc":"Optionally, an init function can be passed, which will be called before the connection is upgraded. It will be passed the Cowboy request, and must return {:ok, context} . The context will then be passed to the topic. This can be useful for authentication: for unauthenticated users, return an error result to prevent the socket getting established; for authenticated users, include the user ID in the context so it can be used for authorisation/identification within a topic.","ref":"cowboy-adapter.html#context","title":"Cowboy adapter - Context","type":"extras"},{"doc":"You can now connect to Topical from a JavaScript client .","ref":"cowboy-adapter.html#client","title":"Cowboy adapter - Client","type":"extras"},{"doc":"You can connect to a Topical server that has been exposed by an adapter, using the JavaScript client. The client is available on npm: npm install @topical/core (Or you may prefer to use the React client - see below.) type ListModel = { items: Record&lt;string, { text: string, done: boolean }&gt;; order: string[]; } // Setup the socket const socket = new Socket(&quot;ws://example.com/socket&quot;); // Subscribe to a topic (returns a function to unsibscribe) const unsubscribe = socket.subscribe&lt;ListModel&gt;( [&quot;lists&quot;, &quot;foo&quot;], (list: ListModel) =&gt; { console.log(value); }, (error) =&gt; { ... } ); // Execute an action const itemId = await socket.execute([&quot;lists&quot;, &quot;foo&quot;], &quot;add_item&quot;, &quot;First item&quot;); // (The subscription should have been updated) // Send a notification socket.notify([&quot;lists&quot;, &quot;foo&quot;], &quot;update_item&quot;, itemId, &quot;Inaugural item&quot;) // (The subscription should have been updated again) // Unsubscribe unsubscribe();","ref":"javascript-client.html","title":"JavaScript client","type":"extras"},{"doc":"Instead of using the JavaScript client directly, you can use the React client. Install it from npm: npm install @topical/react Setup the socket using the provider: import { SocketProvider } from &quot;@topical/react&quot;; function getSocketUrl() { // TODO return &quot;ws://example.com/socket&quot;; } function App() { return ( &lt;SocketProvider url={getSocketUrl()}&gt; // ... &lt;/SocketProvider&gt; ); } Then use the useTopic hook in your components to subscribe to your topic: import { useTopic } from &quot;@topical/react&quot;; function List({ id }) { const [list, { execute, notify }] = useTopic&lt;models.List&gt;(&quot;lists&quot;, id); const addItem = useCallback( (text: string) =&gt; execute(&quot;add_item&quot;, text), [execute] ); if (list) { return ( // ... ); } else { return &lt;p&gt;Loading...&lt;/p&gt; } } If you need access to the underlying socket (or the status), you can use the useSocket hook: import { useSocket } from &quot;@topical/react&quot;; function SocketStatus() { const [_socket, state] = useSocket(); return &lt;p&gt;{state}&lt;/p&gt;; }","ref":"javascript-client.html#react-client","title":"JavaScript client - React client","type":"extras"}]